@@ -8,19 +8,23 @@ Released under the BSD License.
 
 @author      Erki Suurjaak
 @created     23.10.2021
-@modified    14.10.2022
+@modified    05.07.2023
 ------------------------------------------------------------------------------
 """
 ## @namespace grepros.main
 import argparse
 import atexit
 import collections
+import logging
 import os
+import random
 import re
 import sys
 
-from . import __version__, inputs, outputs, search
-from . common import ConsolePrinter, parse_datetime
+import six
+
+from . import __title__, __version__, __version_date__, api, inputs, outputs, search
+from . common import ConsolePrinter, MatchMarkers, parse_datetime
 from . import plugins
 
 
@@ -37,44 +41,44 @@ target matches if any value matches.
 Example usage:
 
 Search for "my text" in all bags under current directory and subdirectories:
-    grepros -r "my text"
+    %(title)s -r "my text"
 
 Print 30 lines of the first message from each live ROS topic:
-    grepros --max-per-topic 1 --lines-per-message 30 --live
+    %(title)s --max-per-topic 1 --lines-per-message 30 --live
 
 Find first message containing "future" (case-sensitive) in my.bag:
-    grepros future -I --max-count 1 --name my.bag
+    %(title)s future -I --max-count 1 --name my.bag
 
 Find 10 messages, from geometry_msgs package, in "map" frame,
 from bags in current directory, reindexing any unindexed bags:
-    grepros frame_id=map --type geometry_msgs/* --max-count 10  --reindex-if-unindexed
+    %(title)s frame_id=map --type geometry_msgs/* --max-count 10  --reindex-if-unindexed
 
 Pipe all diagnostics messages with "CPU usage" from live ROS topics to my.bag:
-    grepros "CPU usage" --type *DiagnosticArray --no-console-output --write my.bag
+    %(title)s "CPU usage" --type *DiagnosticArray --no-console-output --write my.bag
 
 Find messages with field "key" containing "0xA002",
 in topics ending with "diagnostics", in bags under "/tmp":
-    grepros key=0xA002 --topic *diagnostics --path /tmp
+    %(title)s key=0xA002 --topic *diagnostics --path /tmp
 
 Find diagnostics_msgs messages in bags in current directory,
 containing "navigation" in fields "name" or "message",
 print only header stamp and values:
-    grepros --type diagnostic_msgs/* --select-field name message \\
-            --print-field header.stamp status.values -- navigation
+    %(title)s --type diagnostic_msgs/* --select-field name message \\
+            --emit-field header.stamp status.values -- navigation
 
-Print first message from each lidar topic on host 1.2.3.4:
+Print first message from each lidar topic on host 1.2.3.4, without highlight:
     ROS_MASTER_URI=http://1.2.3.4::11311 \\
-    grepros --live --topic *lidar* --max-per-topic 1
+    %(title)s --live --topic *lidar* --max-per-topic 1 --no-highlight
 
 Export all bag messages to SQLite and Postgres, print only export progress:
-    grepros -n my.bag --write my.bag.sqlite --no-console-output --no-verbose --progress
+    %(title)s -n my.bag --write my.bag.sqlite --no-console-output --no-verbose --progress
 
-    grepros -n my.bag --write postgresql://user@host/dbname \\
+    %(title)s -n my.bag --write postgresql://user@host/dbname \\
             --no-console-output --no-verbose --progress
-    """,
+    """ % dict(title=__title__),
 
     "arguments": [
-        dict(args=["PATTERNS"], nargs="*", metavar="PATTERN",
+        dict(args=["PATTERN"], nargs="*", default=[],
              help="pattern(s) to find in message field values,\n"
                   "all messages match if not given,\n"
                   "can specify message field as NAME=PATTERN\n"
@@ -85,7 +89,7 @@ Export all bag messages to SQLite and Postgres, print only export progress:
              help="show this help message and exit"),
 
         dict(args=["-F", "--fixed-strings"],
-             dest="RAW", action="store_true",
+             dest="FIXED_STRING", action="store_true",
              help="PATTERNs are ordinary strings, not regular expressions"),
 
         dict(args=["-I", "--no-ignore-case"],
@@ -94,11 +98,12 @@ Export all bag messages to SQLite and Postgres, print only export progress:
 
         dict(args=["-v", "--invert-match"],
              dest="INVERT", action="store_true",
-             help="select non-matching messages"),
+             help="select messages not matching PATTERN"),
 
         dict(args=["--version"],
              dest="VERSION", action="version",
-             version="grepros (grep for ROS bag files and live topics) %s" % __version__,
+             version="%s: grep for ROS bag files and live topics, v%s (%s)" %
+                     (__title__, __version__, __version_date__),
              help="display version information and exit"),
 
         dict(args=["--live"],
@@ -121,39 +126,43 @@ Export all bag messages to SQLite and Postgres, print only export progress:
              dest="WRITE_OPTIONS", default=argparse.SUPPRESS, help=argparse.SUPPRESS),
 
         dict(args=["--plugin"],
-             dest="PLUGINS", metavar="PLUGIN", nargs="+", default=[], action="append",
+             dest="PLUGIN", nargs="+", default=[], action="append",
              help="load a Python module or class as plugin"),
+
+        dict(args=["--stop-on-error"],
+             dest="STOP_ON_ERROR", action="store_true",
+             help="stop further execution on any error like unknown message type"),
     ],
 
     "groups": {"Filtering": [
 
         dict(args=["-t", "--topic"],
-             dest="TOPICS", metavar="TOPIC", nargs="+", default=[], action="append",
-             help="ROS topics to scan if not all (supports * wildcards)"),
+             dest="TOPIC", nargs="+", default=[], action="append",
+             help="ROS topics to read if not all (supports * wildcards)"),
 
         dict(args=["-nt", "--no-topic"],
-             dest="SKIP_TOPICS", metavar="TOPIC", nargs="+", default=[], action="append",
+             dest="SKIP_TOPIC", metavar="TOPIC", nargs="+", default=[], action="append",
              help="ROS topics to skip (supports * wildcards)"),
 
         dict(args=["-d", "--type"],
-             dest="TYPES", metavar="TYPE", nargs="+", default=[], action="append",
-             help="ROS message types to scan if not all (supports * wildcards)"),
+             dest="TYPE", nargs="+", default=[], action="append",
+             help="ROS message types to read if not all (supports * wildcards)"),
 
         dict(args=["-nd", "--no-type"],
-             dest="SKIP_TYPES", metavar="TYPE", nargs="+", default=[], action="append",
+             dest="SKIP_TYPE", metavar="TYPE", nargs="+", default=[], action="append",
              help="ROS message types to skip (supports * wildcards)"),
 
         dict(args=["--condition"],
-             dest="CONDITIONS", metavar="CONDITION", nargs="+", default=[], action="append",
+             dest="CONDITION", nargs="+", default=[], action="append",
              help="extra conditions to require for matching messages,\n"
                   "as ordinary Python expressions, can refer to last messages\n"
-                  "in topics as {topic /my/topic}; topic name can contain wildcards.\n"
-                  'E.g. --condition "{topic /robot/enabled}.data" matches\n'
+                  "in topics as <topic /my/topic>; topic name can contain wildcards.\n"
+                  'E.g. --condition "<topic /robot/enabled>.data" matches\n'
                   "messages only while last message in '/robot/enabled' has data=true."),
 
         dict(args=["-t0", "--start-time"],
              dest="START_TIME", metavar="TIME",
-             help="earliest timestamp of messages to scan\n"
+             help="earliest timestamp of messages to read\n"
                   "as relative seconds if signed,\n"
                   "or epoch timestamp or ISO datetime\n"
                   "(for bag input, relative to bag start time\n"
@@ -163,7 +172,7 @@ Export all bag messages to SQLite and Postgres, print only export progress:
 
         dict(args=["-t1", "--end-time"],
              dest="END_TIME", metavar="TIME",
-             help="latest timestamp of messages to scan\n"
+             help="latest timestamp of messages to read\n"
                   "as relative seconds if signed,\n"
                   "or epoch timestamp or ISO datetime\n"
                   "(for bag input, relative to bag start time\n"
@@ -183,32 +192,32 @@ Export all bag messages to SQLite and Postgres, print only export progress:
 
         dict(args=["--every-nth-message"],
              dest="NTH_MESSAGE", metavar="NUM", type=int, default=1,
-             help="scan every Nth message within topic"),
+             help="read every Nth message within topic"),
 
         dict(args=["--every-nth-interval"],
              dest="NTH_INTERVAL", metavar="SECONDS", type=int, default=0,
-             help="scan messages at least N seconds apart within topic"),
+             help="read messages at least N seconds apart within topic"),
 
         dict(args=["--every-nth-match"],
              dest="NTH_MATCH", metavar="NUM", type=int, default=1,
              help="emit every Nth match in topic"),
 
         dict(args=["-sf", "--select-field"],
-             dest="SELECT_FIELDS", metavar="FIELD", nargs="+", default=[], action="append",
+             dest="SELECT_FIELD", metavar="FIELD", nargs="+", default=[], action="append",
              help="message fields to use in matching if not all\n"
                   "(supports nested.paths and * wildcards)"),
 
         dict(args=["-ns", "--no-select-field"],
-             dest="NOSELECT_FIELDS", metavar="FIELD", nargs="+", default=[], action="append",
+             dest="NOSELECT_FIELD", metavar="FIELD", nargs="+", default=[], action="append",
              help="message fields to skip in matching\n"
                   "(supports nested.paths and * wildcards)"),
 
         dict(args=["-m", "--max-count"],
-             dest="MAX_MATCHES", metavar="NUM", default=0, type=int,
+             dest="MAX_COUNT", metavar="NUM", default=0, type=int,
              help="number of matched messages to emit (per each file if bag input)"),
 
         dict(args=["--max-per-topic"],
-             dest="MAX_TOPIC_MATCHES", metavar="NUM", default=0, type=int,
+             dest="MAX_PER_TOPIC", metavar="NUM", default=0, type=int,
              help="number of matched messages to emit from each topic\n"
                   "(per each file if bag input)"),
 
@@ -237,53 +246,53 @@ Export all bag messages to SQLite and Postgres, print only export progress:
              help="emit NUM messages of leading and trailing context\n"
                   "around match"),
 
-        dict(args=["-pf", "--print-field"],
-             dest="PRINT_FIELDS", metavar="FIELD", nargs="+", default=[], action="append",
-             help="message fields to print in console output if not all\n"
+        dict(args=["-ef", "--emit-field"],
+             dest="EMIT_FIELD", metavar="FIELD", nargs="+", default=[], action="append",
+             help="message fields to emit in console output if not all\n"
                   "(supports nested.paths and * wildcards)"),
 
-        dict(args=["-np", "--no-print-field"],
-             dest="NOPRINT_FIELDS", metavar="FIELD", nargs="+", default=[], action="append",
+        dict(args=["-nf", "--no-emit-field"],
+             dest="NOEMIT_FIELD", metavar="FIELD", nargs="+", default=[], action="append",
              help="message fields to skip in console output\n"
                   "(supports nested.paths and * wildcards)"),
 
         dict(args=["-mo", "--matched-fields-only"],
              dest="MATCHED_FIELDS_ONLY", action="store_true",
-             help="print only the fields where PATTERNs find a match"),
+             help="emit only the fields where PATTERNs find a match in console output"),
 
         dict(args=["-la", "--lines-around-match"],
              dest="LINES_AROUND_MATCH", metavar="NUM", type=int,
-             help="print only matched fields and NUM message lines\n"
-                  "around match"),
+             help="emit only matched fields and NUM message lines\n"
+                  "around match in console output"),
 
         dict(args=["-lf", "--lines-per-field"],
              dest="MAX_FIELD_LINES", metavar="NUM", type=int,
-             help="maximum number of lines to print per field"),
+             help="maximum number of lines to emit per field in console output"),
 
         dict(args=["-l0", "--start-line"],
              dest="START_LINE", metavar="NUM", type=int,
-             help="message line number to start printing from\n"
+             help="message line number to start emitting from in console output\n"
                   "(1-based if positive, counts back from total if negative)"),
 
         dict(args=["-l1", "--end-line"],
              dest="END_LINE", metavar="NUM", type=int,
-             help="message line number to stop printing at\n"
+             help="message line number to stop emitting at in console output\n"
                   "(1-based if positive, counts back from total if negative)"),
 
         dict(args=["-lm", "--lines-per-message"],
              dest="MAX_MESSAGE_LINES", metavar="NUM", type=int,
-             help="maximum number of lines to print per message"),
+             help="maximum number of lines to emit per message in console output"),
 
         dict(args=["--match-wrapper"],
              dest="MATCH_WRAPPER", metavar="STR", nargs="*",
-             help="string to wrap around matched values,\n"
+             help="string to wrap around matched values in console output,\n"
                   "both sides if one value, start and end if more than one,\n"
                   "or no wrapping if zero values\n"
                   '(default "**" in colorless output)'),
 
         dict(args=["--wrap-width"],
              dest="WRAP_WIDTH", metavar="NUM", type=int,
-             help="character width to wrap message YAML output at,\n"
+             help="character width to wrap message YAML console output at,\n"
                   "0 disables (defaults to detected terminal width)"),
 
         dict(args=["--color"], dest="COLOR",
@@ -296,6 +305,9 @@ Export all bag messages to SQLite and Postgres, print only export progress:
         dict(args=["--no-filename"], dest="LINE_PREFIX", action="store_false",
              help="do not print bag filename prefix on each console message line"),
 
+        dict(args=["--no-highlight"], dest="HIGHLIGHT", action="store_false",
+             help="do not highlight matched values"),
+
         dict(args=["--no-console-output"], dest="CONSOLE", action="store_false",
              help="do not print matches to console"),
 
@@ -313,12 +325,12 @@ Export all bag messages to SQLite and Postgres, print only export progress:
     ], "Bag input control": [
 
         dict(args=["-n", "--filename"],
-             dest="FILES", metavar="FILE", nargs="+", default=[], action="append",
-             help="names of ROS bagfiles to scan if not all in directory\n"
+             dest="FILE", nargs="+", default=[], action="append",
+             help="names of ROS bagfiles to read if not all in directory\n"
                   "(supports * wildcards)"),
 
         dict(args=["-p", "--path"],
-             dest="PATHS", metavar="PATH", nargs="+", default=[], action="append",
+             dest="PATH", nargs="+", default=[], action="append",
              help="paths to scan if not current directory\n"
                   "(supports * wildcards)"),
 
@@ -369,6 +381,9 @@ Export all bag messages to SQLite and Postgres, print only export progress:
     ]},
 }
 
+## List of command-line arguments the program was invoked with
+CLI_ARGS = None
+
 
 class HelpFormatter(argparse.RawTextHelpFormatter):
     """RawTextHelpFormatter returning custom metavar for WRITE."""
@@ -396,22 +411,30 @@ def make_parser():
 
 def process_args(args):
     """
-    Converts or combines arguments where necessary, returns args.
+    Converts or combines arguments where necessary, returns full args.
 
     @param   args  arguments object like argparse.Namespace
     """
+    for arg in sum(ARGUMENTS.get("groups", {}).values(), ARGUMENTS["arguments"][:]):
+        name = arg.get("dest") or arg["args"][0]
+        if "version" != arg.get("action") and argparse.SUPPRESS != arg.get("default") \
+        and "HELP" != name and not hasattr(args, name):
+            value = False if arg.get("store_true") else True if arg.get("store_false") else None
+            setattr(args, name, arg.get("default", value))
+
     if args.CONTEXT:
         args.BEFORE = args.AFTER = args.CONTEXT
 
     # Default to printing metadata for publish/write if no console output
-    args.VERBOSE = False if args.SKIP_VERBOSE else (args.VERBOSE or not args.CONSOLE)
+    args.VERBOSE = False if args.SKIP_VERBOSE else \
+                   (args.VERBOSE or not args.CONSOLE and bool(CLI_ARGS))
 
     # Show progress bar only if no console output
     args.PROGRESS = args.PROGRESS and not args.CONSOLE
 
     # Print filename prefix on each console message line if not single specific file
-    args.LINE_PREFIX = args.LINE_PREFIX and (args.RECURSE or len(args.FILES) != 1
-                                             or args.PATHS or any("*" in x for x in args.FILES))
+    args.LINE_PREFIX = args.LINE_PREFIX and (args.RECURSE or len(args.FILE) != 1
+                                             or args.PATH or any("*" in x for x in args.FILE))
 
     for k, v in vars(args).items():  # Flatten lists of lists and drop duplicates
         if k != "WRITE" and isinstance(v, list):
@@ -420,9 +443,9 @@ def process_args(args):
                               if not (x in here or here.add(x))])
 
     for n, v in [("START_TIME", args.START_TIME), ("END_TIME", args.END_TIME)]:
-        if v is None: continue  # for v, n
+        if not isinstance(v, (six.binary_type, six.text_type)): continue  # for v, n
         try: v = float(v)
-        except Exception: pass
+        except Exception: pass  # If numeric, leave as string for source to process as relative time
         try: not isinstance(v, float) and setattr(args, n, parse_datetime(v))
         except Exception: pass
 
@@ -455,14 +478,14 @@ def validate_args(args):
         except Exception: errors[""].append("Invalid ISO datetime for %s: %s" %
                                             (n.lower().replace("_", " "), v))
 
-    for v in args.PATTERNS if not args.RAW else ():
+    for v in args.PATTERN if not args.FIXED_STRING else ():
         split = v.find("=", 1, -1)  # May be "PATTERN" or "attribute=PATTERN"
         v = v[split + 1:] if split > 0 else v
-        try: re.compile(re.escape(v) if args.RAW else v)
+        try: re.compile(re.escape(v) if args.FIXED_STRING else v)
         except Exception as e:
             errors["Invalid regular expression"].append("'%s': %s" % (v, e))
 
-    for v in args.CONDITIONS:
+    for v in args.CONDITION:
         v = inputs.ConditionMixin.TOPIC_RGX.sub("dummy", v)
         try: compile(v, "", "eval")
         except SyntaxError as e:
@@ -472,11 +495,11 @@ def validate_args(args):
             errors["Invalid condition"].append("'%s': %s" % (v, e))
 
     for err in errors.get("", []):
-        ConsolePrinter.error(err)
+        ConsolePrinter.log(logging.ERROR, err)
     for category in filter(bool, errors):
-        ConsolePrinter.error(category)
+        ConsolePrinter.log(logging.ERROR, category)
         for err in errors[category]:
-            ConsolePrinter.error("  %s" % err)
+            ConsolePrinter.log(logging.ERROR, "  %s" % err)
     return not errors
 
 
@@ -490,27 +513,30 @@ def flush_stdout():
 def preload_plugins():
     """Imports and initializes plugins from auto-load folder and from arguments."""
     plugins.add_write_format("bag", outputs.BagSink, "bag", [
-        ("overwrite=true|false",   "overwrite existing file in bag output\n"
+        ("overwrite=true|false",   "overwrite existing file\nin bag output\n"
                                    "instead of appending to if bag or database\n"
                                    "or appending unique counter to file name\n"
                                    "(default false)")
 
     ])
-    args = make_parser().parse_known_args()[0] if "--plugin" in sys.argv else None
+    args = make_parser().parse_known_args(CLI_ARGS)[0] if "--plugin" in CLI_ARGS else None
     try: plugins.init(process_args(args) if args else None)
     except ImportWarning: sys.exit(1)
 
 
 def run():
     """Parses command-line arguments and runs search."""
+    global CLI_ARGS
+    CLI_ARGS = sys.argv[1:]
+    MatchMarkers.populate("%08x" % random.randint(1, 1E9))
     preload_plugins()
     argparser = make_parser()
-    if len(sys.argv) < 2:
+    if not CLI_ARGS:
         argparser.print_usage()
         return
 
     atexit.register(flush_stdout)
-    args, _ = argparser.parse_known_args()
+    args, _ = argparser.parse_known_args(CLI_ARGS)
     if args.HELP:
         argparser.print_help()
         return
@@ -536,8 +562,8 @@ def run():
 
         thread_excepthook = lambda t, e: (ConsolePrinter.error(t), sys.exit(1))
         source.thread_excepthook = sink.thread_excepthook = thread_excepthook
-        searcher = plugins.load("search", args) or search.Searcher(args)
-        searcher.search(source, sink)
+        grepper = plugins.load("scan", args) or search.Scanner(args)
+        grepper.work(source, sink)
     except BREAK_EXS:
         try: source and source.close()
         except (Exception, KeyboardInterrupt): pass
@@ -550,6 +576,14 @@ def run():
     finally:
         sink and sink.close()
         source and source.close()
+        api.shutdown_node()
+
+
+__all__ = [
+    "ARGUMENTS", "CLI_ARGS", "HelpFormatter",
+    "make_parser", "process_args", "validate_args", "flush_stdout", "preload_plugins", "run",
+]
+
 
 
 if "__main__" == __name__:
