@@ -8,11 +8,15 @@ Released under the BSD License.
 
 @author      Erki Suurjaak
 @created     01.11.2021
-@modified    27.10.2022
+@modified    03.07.2023
 ------------------------------------------------------------------------------
 """
 ## @namespace grepros.ros1
 import collections
+import datetime
+import decimal
+import inspect
+import logging
 import io
 import os
 import shutil
@@ -23,9 +27,12 @@ import genpy.dynamic
 import rosbag
 import roslib
 import rospy
+import six
 
-from . common import ConsolePrinter, MatchMarkers, ProgressBar, format_bytes, memoize
-from . rosapi import TypeMeta, calculate_definition_hash, parse_definition_subtypes
+from . import api
+from . import common
+from . api import TypeMeta, calculate_definition_hash, parse_definition_subtypes
+from . common import ConsolePrinter, memoize
 
 
 ## Bagfile extensions to seek
@@ -53,8 +60,27 @@ SLEEP_INTERVAL = 0.5
 master = None
 
 
-class Bag(rosbag.Bag):
-    """Add message type getters and grepros-specific write() to rosbag.Bag."""
+class ROS1Bag(rosbag.Bag, api.BaseBag):
+    """
+    ROS1 bag reader and writer.
+
+    Extends `rosbag.Bag` with more conveniences, smooths over the rosbag bug of ignoring
+    topic and time filters in format v1.2, and smooths over the rosbag bug
+    of yielding messages of wrong type, if message types in different topics
+    have different packages but identical fields and hashes.
+
+    Does **not** smooth over the rosbag bug of writing different types to one topic.
+
+    rosbag does allow writing messages of different types to one topic,
+    just like live ROS topics can have multiple message types published
+    to one topic. And their serialized bytes will actually be in the bag,
+    but rosbag will only register the first type for this topic (unless it is
+    explicitly given another connection header with metadata on the other type).
+
+    All messages yielded will be deserialized by rosbag as that first type,
+    and whether reading will raise an exception or not depends on whether
+    the other type has enough bytes to be deserialized as that first type.
+    """
 
     # {(typename, typehash): message type class}
     __TYPES    = {}
@@ -68,25 +94,43 @@ class Bag(rosbag.Bag):
 
     def __init__(self, *args, **kwargs):
         """
+        @param   f           bag file path, or a stream object
         @param   mode        mode to open bag in, defaults to "r" (read mode)
         @param   reindex     if true and bag is unindexed, make a copy
                              of the file (unless unindexed format) and reindex original
         @param   progress    show progress bar with reindexing status
+        @param   kwargs      additional keyword arguments for `rosbag.Bag`, like `compression`
         """
+        self.__topics = {}    # {(topic, typename, typehash): message count}
+        self.__iterer = None  # Generator from read_messages() for next()
+        f,    args = (args[0] if args else kwargs.pop("f")), args[1:]
+        mode, args = (args[0] if args else kwargs.pop("mode", "r")), args[1:]
+        if mode not in self.MODES: raise ValueError("invalid mode %r" % mode)
+
         kwargs.setdefault("skip_index", True)
         reindex, progress = (kwargs.pop(k, False) for k in ("reindex", "progress"))
-        filename, args = (args[0] if args else kwargs.pop("f")), args[1:]
-        mode,     args = (args[0] if args else kwargs.pop("mode", "r")), args[1:]
+        getargspec = getattr(inspect, "getfullargspec", inspect.getargspec)
+        for n in set(kwargs) - set(getargspec(rosbag.Bag).args): kwargs.pop(n)
+
+        if common.is_stream(f):
+            if not common.verify_io(f, mode):
+                raise io.UnsupportedOperation({"r": "read", "w": "write", "a": "append"}[mode])
+            super(ROS1Bag, self).__init__(f, mode, *args, **kwargs)
+            self._populate_meta()
+            return
+
+        f = str(f)
+        if "a" == mode and (not os.path.exists(f) or not os.path.getsize(f)):
+            mode = "w"  # rosbag raises error on append if no file or empty file
+            os.path.exists(f) and os.remove(f)
+
         try:
-            super(Bag, self).__init__(filename, mode, *args, **kwargs)
+            super(ROS1Bag, self).__init__(f, mode, *args, **kwargs)
         except rosbag.ROSBagUnindexedException:
             if not reindex: raise
-            Bag.reindex_file(filename, progress, *args, **kwargs)
-            super(Bag, self).__init__(filename, mode, *args, **kwargs)
-
-        self.__topics = {}  # {(topic, typename, typehash): message count}
-
-        self.__populate_meta()
+            Bag.reindex_file(f, progress, *args, **kwargs)
+            super(ROS1Bag, self).__init__(f, mode, *args, **kwargs)
+        self._populate_meta()
 
 
     def get_message_definition(self, msg_or_type):
@@ -105,63 +149,93 @@ class Bag(rosbag.Bag):
 
         @param   typehash  message type definition hash, if any
         """
-        self.__ensure_typedef(typename, typehash)
+        if (typename, typehash) not in self.__TYPES: self._ensure_typedef(typename, typehash)
         typehash = typehash or next((h for n, h in self.__TYPEDEFS if n == typename), None)
         typekey = (typename, typehash)
         if typekey not in self.__TYPES and typekey in self.__TYPEDEFS:
             for n, c in genpy.dynamic.generate_dynamic(typename, self.__TYPEDEFS[typekey]).items():
                 self.__TYPES[(n, c._md5sum)] = c
-        return self.__TYPES.get(typekey) or get_message_class(typename)
+        return self.__TYPES.get(typekey)
 
 
     def get_message_type_hash(self, msg_or_type):
-        """Returns ROS1 message type MD5 hash."""
+        """Returns ROS1 message type MD5 hash, or None if unknown type."""
         if is_ros_message(msg_or_type): return msg_or_type._md5sum
         typename = msg_or_type
         typehash = next((h for n, h in self.__TYPEDEFS if n == typename), None)
         if not typehash:
-            self.__ensure_typedef(typename)
+            self._ensure_typedef(typename)
             typehash = next((h for n, h in self.__TYPEDEFS if n == typename), None)
-        return typehash or get_message_type_hash(typename)
+        return typehash
 
 
-    def get_qoses(self, topic, typename):
-        """Returns None."""
-        return None
+    def get_start_time(self):
+        """Returns the start time of the bag, as UNIX timestamp, or None if bag empty."""
+        try: return super(ROS1Bag, self).get_start_time()
+        except Exception: return None
 
 
-    def get_topic_info(self):
+    def get_end_time(self):
+        """Returns the end time of the bag, as UNIX timestamp, or None if bag empty."""
+        try: return super(ROS1Bag, self).get_end_time()
+        except Exception: return None
+
+
+    def get_topic_info(self, *_, **__):
         """Returns topic and message type metainfo as {(topic, typename, typehash): count}."""
         return dict(self.__topics)
 
 
-    def read_messages(self, topics=None, start_time=None, end_time=None, connection_filter=None, raw=False):
+    def read_messages(self, topics=None, start_time=None, end_time=None,
+                      raw=False, connection_filter=None, **__):
         """
         Yields messages from the bag, optionally filtered by topic, timestamp and connection details.
 
         @param   topics             list of topics or a single topic.
                                     If an empty list is given, all topics will be read.
-        @param   start_time         earliest timestamp of messages to return
-        @param   end_time           latest timestamp of messages to return
+        @param   start_time         earliest timestamp of messages to return,
+                                    as ROS time or convertible (int/float/duration/datetime/decimal)
+        @param   end_time           latest timestamp of messages to return,
+                                    as ROS time or convertible (int/float/duration/datetime/decimal)
         @param   connection_filter  function to filter connections to include
-        @param   raw                if True, then generate tuples of
-                                    (datatype, (data, md5sum, position), pytype)
-        @return                     generator of (topic, message, timestamp) tuples
+        @param   raw                if true, then returned messages are tuples of
+                                    (typename, bytes, typehash, typeclass)
+                                    or (typename, bytes, typehash, position, typeclass),
+                                    depending on file format
+        @return                     BagMessage namedtuples of
+                                    (topic, message, timestamp as rospy.Time)
         """
+        if self.closed: raise ValueError("I/O operation on closed file.")
+        if "w" == self._mode: raise io.UnsupportedOperation("read")
+
         hashtypes = {}
         for n, h in self.__TYPEDEFS: hashtypes.setdefault(h, []).append(n)
-        read_topics = topics if isinstance(topics, list) else [topics] if topics else None
+        read_topics = topics if isinstance(topics, (dict, list, set, tuple)) else \
+                      [topics] if topics else None
         dupes = {t: (n, h) for t, n, h in self.__topics
                  if (read_topics is None or t in read_topics) and len(hashtypes.get(h, [])) > 1}
 
+        # Workaround for rosbag.Bag ignoring topic and time filters in format v1.2
+        if self.version != 102 or (not topics and start_time is None and end_time is None):
+            in_range = lambda *_: True
+        else: in_range = lambda t, s: ((not read_topics or t in read_topics) and
+                                       (start_time is None or s >= start_time) and
+                                       (end_time is None or s <= end_time))
+
+        start_time, end_time = map(to_time, (start_time, end_time))
         kwargs = dict(topics=topics, start_time=start_time, end_time=end_time,
                       connection_filter=connection_filter, raw=raw)
         if not dupes:
-            for topic, msg, stamp in super(Bag, self).read_messages(**kwargs):
-                yield topic, msg, stamp
+            for topic, msg, stamp in super(ROS1Bag, self).read_messages(**kwargs):
+                if in_range(topic, stamp):
+                    TypeMeta.make(msg, topic, self)
+                    yield self.BagMessage(topic, msg, stamp)
+                if self.closed: break  # for topic
             return
 
-        for topic, msg, stamp in super(Bag, self).read_messages(**kwargs):
+        for topic, msg, stamp in super(ROS1Bag, self).read_messages(**kwargs):
+            if not in_range(topic, stamp): continue  # for topic
+
             # Workaround for rosbag bug of using wrong type for identical type hashes
             if topic in dupes:
                 typename, typehash = (msg[0], msg[2]) if raw else (msg._type, msg._md5sum)
@@ -169,28 +243,84 @@ class Bag(rosbag.Bag):
                     if raw:
                         msg = msg[:-1] + (self.get_message_class(typename, typehash), )
                     else:
-                        msg = self.__convert_message(msg, *dupes[topic])
-            yield topic, msg, stamp
+                        msg = self._convert_message(msg, *dupes[topic])
+            TypeMeta.make(msg, topic, self)
+            yield self.BagMessage(topic, msg, stamp)
+            if self.closed: break  # for topic
 
 
-    def write(self, topic, msg, stamp, *_, **__):
-        """Writes a message to the bag."""
-        return super(Bag, self).write(topic, msg, stamp)
+    def write(self, topic, msg, t=None, raw=False, connection_header=None, **__):
+        """
+        Writes a message to the bag.
+
+        Populates connection header if topic already in bag but with a different message type.
+
+        @param   topic              name of topic
+        @param   msg                ROS1 message
+        @param   t                  message timestamp if not using current wall time,
+                                    as ROS time or convertible (int/float/duration/datetime/decimal)
+        @param   raw                if true, `msg` is in raw format,
+                                    (typename, bytes, typehash, typeclass)
+        @param   connection_header  custom connection record for topic,
+                                    as {"topic", "type", "md5sum", "message_definition"}
+        """
+        if self.closed: raise ValueError("I/O operation on closed file.")
+        if "r" == self._mode: raise io.UnsupportedOperation("write")
+        self._register_write(topic, msg, raw, connection_header)
+        return super(ROS1Bag, self).write(topic, msg, to_time(t), raw, connection_header)
 
 
-    def __convert_message(self, msg, typename2, typehash2=None):
+    def open(self):
+        """Opens the bag file if not already open."""
+        if not self._file:
+            self._open(self.filename, self.mode, allow_unindexed=True)
+
+
+    def close(self):
+        """Closes the bag file."""
+        if self._file:
+            super(ROS1Bag, self).close()
+            self._iterer = None
+            self._clear_index()
+
+
+    def __contains__(self, key):
+        """Returns whether bag contains given topic."""
+        return any(key == t for t, _, _ in self.__topics)
+
+
+    def __next__(self):
+        """Retrieves next message from bag as (topic, message, timestamp)."""
+        if self.closed: raise ValueError("I/O operation on closed file.")
+        if self.__iterer is None: self.__iterer = self.read_messages()
+        return next(self.__iterer)
+
+
+    @property
+    def topics(self):
+        """Returns the list of topics in bag, in alphabetic order."""
+        return sorted((t for t, _, _ in self.__topics), key=str.lower)
+
+
+    @property
+    def closed(self):
+        """Returns whether file is closed."""
+        return not self._file
+
+
+    def _convert_message(self, msg, typename2, typehash2=None):
         """Returns message converted to given type; fields must match."""
         msg2 = self.get_message_class(typename2, typehash2)()
         fields2 = get_message_fields(msg2)
         for fname, ftypename in get_message_fields(msg).items():
             v1 = v2 = getattr(msg, fname)
             if ftypename != fields2.get(fname, ftypename):
-                v2 = self.__convert_message(v1, fields2[fname])
+                v2 = self._convert_message(v1, fields2[fname])
             setattr(msg2, fname, v2)
         return msg2
 
 
-    def __populate_meta(self):
+    def _populate_meta(self):
         """Populates topics and message type definitions and hashes."""
         result = collections.Counter()  # {(topic, typename, typehash): count}
         counts = collections.Counter()  # {connection ID: count}
@@ -203,7 +333,29 @@ class Bag(rosbag.Bag):
         self.__topics = dict(result)
 
 
-    def __ensure_typedef(self, typename, typehash=None):
+    def _register_write(self, topic, msg, raw=False, connection_header=None):
+        """Registers message in local metadata, writes connection header if new type for topic."""
+        if raw: typename, typehash, typeclass = msg[0], msg[2], msg[3]
+        else:   typename, typehash, typeclass = msg._type, msg._md5sum, type(msg)
+        topickey, typekey = (topic, typename, typehash), (typename, typehash)
+
+        if topickey not in self.__topics \
+        and any(topic == t and typekey != (n, h) for t, n, h in self.__topics):
+            # Write connection header if topic already present but with different type
+            if not connection_header:
+                connection_header = {"topic": topic, "type": typename, "md5sum": typehash,
+                                     "message_definition": typeclass._full_text}
+            conn_id, bagself = len(self._connections), super(ROS1Bag, self)
+            connection_info = rosbag.bag._ConnectionInfo(conn_id, topic, connection_header)
+            bagself._write_connection_record(connection_info, encrypt=False)
+            bagself._connections[conn_id] = bagself._topic_connections[topic] = connection_info
+
+        self.__TYPES.setdefault(typekey, typeclass)
+        self.__TYPEDEFS.setdefault(typekey, typeclass._full_text)
+        self.__topics[topickey] = self.__topics.get(topickey, 0) + 1
+
+
+    def _ensure_typedef(self, typename, typehash=None):
         """Parses subtype definition from any full definition where available, if not loaded."""
         typehash = typehash or next((h for n, h in self.__TYPEDEFS if n == typename), None)
         typekey = (typename, typehash)
@@ -234,10 +386,10 @@ class Bag(rosbag.Bag):
         kwargs.update(zip(KWS, args), allow_unindexed=True)
         copied, bar, f2 = False, None, None
         if progress:
-            fmt = lambda s: format_bytes(s, strip=False)
+            fmt = lambda s: common.format_bytes(s, strip=False)
             name, size = os.path.basename(f), os.path.getsize(f)
             aftertemplate = " Reindexing %s (%s): {afterword}" % (name, fmt(size))
-            bar = ProgressBar(size, interval=0.1, pulse=True, aftertemplate=aftertemplate)
+            bar = common.ProgressBar(size, interval=0.1, pulse=True, aftertemplate=aftertemplate)
 
         ConsolePrinter.warn("Unindexed bag %s, reindexing.", f)
         bar and bar.update(0).start()  # Start progress pulse
@@ -272,7 +424,7 @@ class Bag(rosbag.Bag):
         update_bar = noop = lambda s: None
         indexbag, writebag = (inbag, outbag) if inbag.version == 102 else (outbag, None)
         if bar:
-            fmt = lambda s: format_bytes(s, strip=False)
+            fmt = lambda s: common.format_bytes(s, strip=False)
             update_bar = lambda s: (setattr(bar, "afterword", fmt(s)),
                                     setattr(bar, "pulse", False), bar.update(s).stop())
         # v102: build index from inbag, write all messages to outbag.
@@ -287,6 +439,7 @@ class Bag(rosbag.Bag):
         for (topic, msg, t, header) in indexbag.read_messages(return_connection_header=True):
             writebag.write(topic, msg, t, connection_header=header)
             progress(indexbag._file.tell())
+Bag = ROS1Bag
 
 
 
@@ -304,13 +457,16 @@ def init_node(name):
         master = rospy.client.get_master()
         available = None
         while not available:
-            try: master.getSystemState()
+            try: uri = master.getUri()[-1]
             except Exception:
                 if available is None:
-                    ConsolePrinter.error("Unable to register with master. Will keep trying.")
+                    ConsolePrinter.log(logging.ERROR,
+                                       "Unable to register with master. Will keep trying.")
                 available = False
                 time.sleep(SLEEP_INTERVAL)
-            else: available = True
+            else:
+                ConsolePrinter.debug("Connected to ROS master at %s.", uri)
+                available = True
     try: rospy.get_rostime()
     except Exception:  # Init node only if not already inited
         rospy.init_node(name, anonymous=True, disable_signals=True)
@@ -341,6 +497,20 @@ def validate(live=False):
     return not missing
 
 
+@memoize
+def canonical(typename, unbounded=False):
+    """
+    Returns "pkg/Type" for "pkg/subdir/Type".
+
+    @param  unbounded  drop array bounds, e.g. returning "uint8[]" for "uint8[10]"
+    """
+    if typename and typename.count("/") > 1:
+        typename = "%s/%s" % tuple((x[0], x[-1]) for x in [typename.split("/")])[0]
+    if unbounded and typename and "[" in typename:
+        typename = typename[:typename.index("[")] + "[]"
+    return typename
+
+
 def create_publisher(topic, cls_or_typename, queue_size):
     """Returns a rospy.Publisher."""
     def pub_unregister():
@@ -348,7 +518,7 @@ def create_publisher(topic, cls_or_typename, queue_size):
         if not pub.get_num_connections(): super(rospy.Publisher, pub).unregister()
 
     cls = cls_or_typename
-    if isinstance(cls, str): cls = get_message_class(cls)
+    if isinstance(cls, common.TEXT_TYPES): cls = get_message_class(cls)
     pub = rospy.Publisher(topic, cls, queue_size=queue_size)
     pub.unregister = pub_unregister
     return pub
@@ -400,47 +570,56 @@ def format_message_value(msg, name, value):
     if not isinstance(msg, genpy.TVal) or name not in LENS:
         return v
 
-    EXTRA = sum(v.count(x) * len(x) for x in (MatchMarkers.START, MatchMarkers.END))
+    EXTRA = sum(v.count(x) * len(x) for x in (common.MatchMarkers.START, common.MatchMarkers.END))
     return ("%%%ds" % (LENS[name] + EXTRA)) % v  # Default %10s/%9s for secs/nsecs
 
 
 @memoize
 def get_message_class(typename):
     """Returns ROS1 message class."""
-    return roslib.message.get_message_class(typename)
-
-
-def get_message_data(msg):
-    """Returns ROS1 message as a serialized binary."""
-    with TypeMeta.make(msg) as m:
-        if m.data is not None: return m.data
-    buf = io.BytesIO()
-    msg.serialize(buf)
-    return buf.getvalue()
+    if typename in ROS_TIME_TYPES:
+        return next(k for k, v in ROS_TIME_CLASSES.items() if v == typename)
+    if typename in ("genpy/Time", "rospy/Time"):
+        return rospy.Time
+    if typename in ("genpy/Duration", "rospy/Duration"):
+        return rospy.Duration
+    try: return roslib.message.get_message_class(typename)
+    except Exception: return None
 
 
 def get_message_definition(msg_or_type):
-    """Returns ROS1 message type definition full text, including subtype definitions."""
+    """
+    Returns ROS1 message type definition full text, including subtype definitions.
+
+    Returns None if unknown type.
+    """
     msg_or_cls = msg_or_type if is_ros_message(msg_or_type) else get_message_class(msg_or_type)
-    return msg_or_cls._full_text
+    return getattr(msg_or_cls, "_full_text", None)
 
 
 def get_message_type_hash(msg_or_type):
-    """Returns ROS message type MD5 hash."""
+    """Returns ROS message type MD5 hash, or "" if unknown type."""
     msg_or_cls = msg_or_type if is_ros_message(msg_or_type) else get_message_class(msg_or_type)
-    return msg_or_cls._md5sum
+    return getattr(msg_or_cls, "_md5sum", "")
 
 
 def get_message_fields(val):
-    """Returns OrderedDict({field name: field type name}) if ROS1 message, else {}."""
-    names = getattr(val, "__slots__", [])
-    if isinstance(val, tuple(ROS_TIME_CLASSES)):  # Empty __slots__
-        names = genpy.TVal.__slots__
-    return collections.OrderedDict(zip(names, getattr(val, "_slot_types", [])))
+    """
+    Returns OrderedDict({field name: field type name}) if ROS1 message, else {}.
+
+    @param   val  ROS1 message class or instance
+    """
+    names, types = (getattr(val, k, []) for k in ("__slots__", "_slot_types"))
+    # Bug in genpy: class slot types defined as "int32", but everywhere else types use "uint32"
+    if isinstance(val, genpy.TVal): names, types = genpy.TVal.__slots__, ["uint32", "uint32"]
+    return collections.OrderedDict(zip(names, types))
 
 
 def get_message_type(msg_or_cls):
     """Returns ROS1 message type name, like "std_msgs/Header"."""
+    if is_ros_time(msg_or_cls):
+        cls = msg_or_cls if inspect.isclass(msg_or_cls) else type(msg_or_cls)
+        return "duration" if "duration" in cls.__name__.lower() else "time"
     return msg_or_cls._type
 
 
@@ -450,9 +629,16 @@ def get_message_value(msg, name, typename):
     return list(v) if typename.startswith("uint8[") and isinstance(v, bytes) else v
 
 
-def get_rostime():
-    """Returns current ROS1 time, as rospy.Time."""
-    return rospy.get_rostime()
+def get_rostime(fallback=False):
+    """
+    Returns current ROS1 time, as rospy.Time.
+
+    @param   fallback  use wall time if node not initialized
+    """
+    try: return rospy.get_rostime()
+    except Exception:
+        if fallback: return make_time(time.time())
+        raise
 
 
 def get_topic_types():
@@ -472,16 +658,17 @@ def get_topic_types():
 
 def is_ros_message(val, ignore_time=False):
     """
-    Returns whether value is a ROS1 message or special like ROS1 time/duration.
+    Returns whether value is a ROS1 message or special like ROS1 time/duration class or instance.
 
     @param  ignore_time  whether to ignore ROS1 time/duration types
     """
-    return isinstance(val, genpy.Message if ignore_time else (genpy.Message, genpy.TVal))
+    isfunc = issubclass if inspect.isclass(val) else isinstance
+    return isfunc(val, genpy.Message if ignore_time else (genpy.Message, genpy.TVal))
 
 
 def is_ros_time(val):
-    """Returns whether value is a ROS1 time/duration."""
-    return isinstance(val, genpy.TVal)
+    """Returns whether value is a ROS1 time/duration class or instance."""
+    return issubclass(val, genpy.TVal) if inspect.isclass(val) else isinstance(val, genpy.TVal)
 
 
 def make_duration(secs=0, nsecs=0):
@@ -494,6 +681,22 @@ def make_time(secs=0, nsecs=0):
     return rospy.Time(secs=secs, nsecs=nsecs)
 
 
+def serialize_message(msg):
+    """Returns ROS1 message as a serialized binary."""
+    with TypeMeta.make(msg) as m:
+        if m.data is not None: return m.data
+    buf = io.BytesIO()
+    msg.serialize(buf)
+    return buf.getvalue()
+
+
+def deserialize_message(raw, cls_or_typename):
+    """Returns ROS1 message or service request/response instantiated from serialized binary."""
+    cls = cls_or_typename
+    if isinstance(cls, common.TEXT_TYPES): cls = get_message_class(cls)
+    return cls().deserialize(raw)
+
+
 @memoize
 def scalar(typename):
     """
@@ -501,7 +704,7 @@ def scalar(typename):
 
     Returns type unchanged if already a scalar.
     """
-    return typename[:typename.index("[")] if "[" in typename else typename
+    return typename[:typename.index("[")] if typename and "[" in typename else typename
 
 
 def set_message_value(obj, name, value):
@@ -522,3 +725,29 @@ def to_sec(val):
 def to_sec_nsec(val):
     """Returns value as (seconds, nanoseconds) if value is ROS1 time/duration, else value."""
     return (val.secs, val.nsecs) if isinstance(val, genpy.TVal) else val
+
+
+def to_time(val):
+    """Returns value as ROS1 time if convertible (int/float/duration/datetime/decimal), else value."""
+    result = val
+    if isinstance(val, decimal.Decimal):
+        result = rospy.Time(int(val), float(val % 1) * 10**9)
+    elif isinstance(val, datetime.datetime):
+        result = rospy.Time(int(val.timestamp()), 1000 * val.microsecond)
+    elif isinstance(val, six.integer_types + (float, )):
+        result = rospy.Time(val)
+    elif isinstance(val, genpy.Duration):
+        result = rospy.Time(val.secs, val.nsecs)
+    return result
+
+
+__all__ = [
+    "BAG_EXTENSIONS", "ROS_ALIAS_TYPES", "ROS_TIME_CLASSES", "ROS_TIME_TYPES", "SKIP_EXTENSIONS",
+    "SLEEP_INTERVAL", "TYPECLASSES", "Bag", "ROS1Bag", "master",
+    "canonical", "create_publisher", "create_subscriber", "deserialize_message",
+    "format_message_value", "get_message_class", "get_message_definition", "get_message_fields",
+    "get_message_type", "get_message_type_hash", "get_message_value", "get_rostime",
+    "get_topic_types", "init_node", "is_ros_message", "is_ros_time", "make_duration", "make_time",
+    "scalar", "serialize_message", "set_message_value", "shutdown_node", "to_nsec", "to_sec",
+    "to_sec_nsec", "to_time", "validate",
+]
